#define OS2VERSION 20      /* Must be 20 for 32-bit EPM */
#define INCL_DOSFILEMGR   /* File Manager values */
#include <EDLL.h>

#include <stdio.h>  /* Needed for sprintf() prototype */

#define DEFAULT_HIGHLIGHT_COLOR  ((DOSCOLOR_MAGENTA << 4) + DOSCOLOR_YELLOW)

   /* lstrcpy() copies one LSTRING to another */
#if 0  /* Lowest overhead, but no type-checking */
  #define lstrcpy(dst, src) memcpy((dst), (src), sizeofLSTRING((src)->lsLength))
#else  /* Let's stick with the safer version... */
  _Inline VOID lstrcpy(PLSTRING dst, PLSTRING src) {
     memcpy(dst, src, sizeofLSTRING(src->lsLength));
  }
#endif


typedef struct _save_pos {
   ULONG  line;
   ULONG  col;
   ULONG  cx;
   ULONG  cy;
} SAVE_POS, * PSAVE_POS;

   /* Forward declarations */
BOOL is_substr(PSZ word, PLSTRING line);
VOID lstrupcase( PLSTRING s );
VOID save_pos( HWND, PSAVE_POS );
VOID restore_pos( HWND, PSAVE_POS );
BOOL _Optlink lstreq_test( const PLSTRING s1, const PLSTRING s2);

/* Execute a given command against each file in the ring. */
ULONG EXPENTRY Each( HWND hwndClient, PSZ command) {
   ULONG start_fid, end_fid;
   ULONG curr_fid = (ULONG)-1;
   BOOL is_visible;
   SHORT rc;

   EtkQueryFileID( hwndClient, &start_fid);
   end_fid = start_fid;
   EtkQueryFileField( hwndClient, VISIBLE_FIELD, 0 /*Fileid*/, (PLONG)&is_visible);
   /* NB:  If initial file is not visible, we'll never get back to it,  */
   /* and we'll loop forever.  Not a good idea.  So, we check the       */
   /* .visible flag of the first file.                                  */
   do {
      rc = EtkExecuteCommand( hwndClient, command);
      #ifdef DEBUG
       {
         unsigned char Text[256];
         PLSTRING pFilename;
         EtkQueryFileField( hwndClient, FILENAME_FIELD, 0 /*Fileid*/, (PLONG)&pFilename);
         sprintf(Text, "Got rc %ld for file \"%.*s\"", rc, pFilename->lsLength, pFilename->Data);
         WinMessageBox(HWND_DESKTOP,
                       hwndClient,
                       Text,
                       "EPMCSAMP - Each",
                       0L,
                       MB_OK + MB_MOVEABLE);
       }
      #endif
      EtkProcessEditKey( hwndClient, ETK_NEXT_FILE, 1);
      /* If first file was not visible, and we've already been through */
      /* the loop once, save the "curr_file" we got the last time as   */
      /* the new "end_file".  Then set the "is_visible" flag to 1      */
      /* so that we don't bother with this stuff again.                */
      if ((!is_visible) && curr_fid != (ULONG)-1) {
         end_fid = curr_fid;
         is_visible = 1;
      } /* endif */
      EtkQueryFileID( hwndClient, &curr_fid);
   } while ( end_fid != curr_fid ); /* enddo */
   /* Force activation of original fileid, in case it was a hidden file. */
   WinSendMsg(hwndClient, EPM_EDIT_ACTIVATEFILEID, MPFROMLONG(start_fid), 0);
}

/* Query the timestamp of the current or specified file. */
ULONG EXPENTRY QTS( HWND hwndClient, PSZ filename) {
   UCHAR         zFileName[261];
   UCHAR         edit_cmd[200];
   PSZ           fn_ptr;
   HDIR          FindHandle;
   FILEFINDBUF3  FindBuffer;
   ULONG         FindCount;
   APIRET        rc;           /* Return code */

   if (strlen(filename)) {
      fn_ptr = filename;
   } else {
      PLSTRING lstr_ptr;
      EtkQueryFileField( hwndClient, FILENAME_FIELD, 0 /*Fileid*/, (PLONG)&lstr_ptr);
      memcpy(zFileName, lstr_ptr->Data, lstr_ptr->lsLength);
      zFileName[lstr_ptr->lsLength] = '\0';
      fn_ptr = zFileName;
   } /* endif */

   FindHandle = 0x0001;
   FindCount = 1;

   rc = DosFindFirst(fn_ptr,                /* File pattern */
                     &FindHandle,           /* Directory search handle */
                     0,                     /* Search attribute */
                     (PVOID) &FindBuffer,   /* Result buffer */
                     sizeof(FindBuffer),    /* Result buffer length */
                     &FindCount,            /* # of entries to find */
                     FIL_STANDARD);         /* Return level 1 file info */

   if (rc != 0) {
      sprintf(edit_cmd, "sayerror DosFindFirst error: return code = %ld", rc);
   } else if (!FindBuffer.fdateCreation.month) { /* FAT file system */
      sprintf(edit_cmd, "sayerror Last written %04hu/%02hu/%02hu %02hu:%02hu:%02hu",
              FindBuffer.fdateLastWrite.year + 1980,
              FindBuffer.fdateLastWrite.month,
              FindBuffer.fdateLastWrite.day,
              FindBuffer.ftimeLastWrite.hours,
              FindBuffer.ftimeLastWrite.minutes,
              FindBuffer.ftimeLastWrite.twosecs << 1);
   } else {
      sprintf(edit_cmd, "sayerror Created %04hu/%02hu/%02hu %02hu:%02hu:%02hu  Last accessed %04hu/%02hu/%02hu %02hu:%02hu:%02hu  Last written %04hu/%02hu/%02hu %02hu:%02hu:%02hu",
              FindBuffer.fdateCreation.year + 1980,
              FindBuffer.fdateCreation.month,
              FindBuffer.fdateCreation.day,
              FindBuffer.ftimeCreation.hours,
              FindBuffer.ftimeCreation.minutes,
              FindBuffer.ftimeCreation.twosecs << 1,
              FindBuffer.fdateLastAccess.year + 1980,
              FindBuffer.fdateLastAccess.month,
              FindBuffer.fdateLastAccess.day,
              FindBuffer.ftimeLastAccess.hours,
              FindBuffer.ftimeLastAccess.minutes,
              FindBuffer.ftimeLastAccess.twosecs << 1,
              FindBuffer.fdateLastWrite.year + 1980,
              FindBuffer.fdateLastWrite.month,
              FindBuffer.fdateLastWrite.day,
              FindBuffer.ftimeLastWrite.hours,
              FindBuffer.ftimeLastWrite.minutes,
              FindBuffer.ftimeLastWrite.twosecs << 1);
   } /* endif */

   EtkExecuteCommand( hwndClient, edit_cmd);
}

ULONG FindAllX( HWND hwndClient, PSZ command, BOOL case_insensitive);

/* Find the next occurrence of all of several words.  Case sensitive. */
ULONG EXPENTRY FindAll( HWND hwndClient, PSZ command) {
   return FindAllX( hwndClient, command, 0);
}
/* Find the next occurrence of all of several words.  Case-insensitive. */
ULONG EXPENTRY FindAllC( HWND hwndClient, PSZ command) {
   return FindAllX( hwndClient, command, 1);
}

/* Find the next occurrence of all of several words.  */
ULONG FindAllX( HWND hwndClient, PSZ command, BOOL case_insensitive) {
   SHORT rc;
   UCHAR locate_cmd[1600];
   int   i, num_words=0;
   PSZ   words[256];
   ULONG linenum;
   PLSTRING getText;
   ATTRIBRECTYPE * getAttrs;
   ATTRIBRECTYPE * getALAttr;
   LSTRING local_copy;
   BOOL success;

   words[0] = (PSZ)strtok(command, " ");
   if (!words[0]) {
      return 0;  /* No words?  Nothing to find... */
   } /* endif */

   strcpy(locate_cmd, "xcom l \1");  /* Let's hope that ASCII 1 is a safe delimiter */
   strcat(locate_cmd, words[0]);
   if (case_insensitive) {
      strcat(locate_cmd, "\1c");
   } /* endif */

   /* We now overwrite word[0] ...  That's fine. */
   while (words[num_words]=(PSZ)strtok(NULL, " ")) {
      if (case_insensitive)
         strupr(words[num_words]);
      num_words++;
   } /* endwhile */

   while (1) {
      /* First do the locate, to find a line containing the first string. */
      if (rc=EtkExecuteCommand( hwndClient, locate_cmd))
         return rc;  /* Error; probably -273 == "String not found" */

      EtkQueryFileField( hwndClient, LINE_FIELD, 0 /*Fileid*/, (PLONG)&linenum);
      EtkQueryText( hwndClient, 0 /*Fileid*/, linenum, &getText, &getAttrs, &getALAttr);
      if (case_insensitive) {
         /* Must make a copy, because getText points to the actual text line */
         /* in the file buffer.  Can't be mucking with that!                 */
         lstrcpy(&local_copy, getText);
         getText = &local_copy;
         lstrupcase(getText);
      } /* endif */
      for (i=num_words, success=1; i--; ) {      /* For each word after the first, */
         if (!is_substr(words[i], getText)) {    /* if word not in this line,      */
            success = 0;                         /* failure.                       */
            break;                               /* So don't check the next word.  */
         } /* endif */
      } /* endfor */
      if (success) {                             /* Did we make it through all?    */
         return 0;                               /* Hurray - success!              */
      } /* endif */
      EtkProcessEditKey( hwndClient, ETK_END_LINE, 1);  /* Move to end of line for next search. */
   } /* endwhile */
}

/* Delete all lines containing a given string.  The Locate delimiters */
/* and any desired options are passed as part of the argument string. */
ULONG EXPENTRY Expunge( HWND hwndClient, PSZ locate_arg) {
   UCHAR locate_cmd[1600];
   ULONG count = 0;
   SHORT rc;
   SAVE_POS savepos;

   save_pos( hwndClient, &savepos );
   strcpy(locate_cmd, "xcom l ");
   strcat(locate_cmd, locate_arg);
   if (rc=EtkExecuteCommand( hwndClient, locate_cmd))
      return rc;  /* Error; probably -273 == "String not found" */
   do {
      EtkProcessEditKey( hwndClient, ETK_DELETE_LINE, 1);
      count++;
      EtkProcessEditKey( hwndClient, ETK_BEGIN_LINE, 1);
   } while ( !EtkExecuteCommand( hwndClient, locate_cmd) ); /* enddo */
   restore_pos( hwndClient, &savepos );
   sprintf(locate_cmd, "sayerror Deleted %ld instance%s.", count, count==1 ? "" : "s");
   EtkExecuteCommand( hwndClient, locate_cmd);
}

/* Save a file w/o modifying the date. */
ULONG EXPENTRY SaveOld( HWND hwndClient, PSZ filename) {
   UCHAR         zFileName[261];
   UCHAR         edit_cmd[200];
   PSZ           fn_ptr;
   HDIR          FindHandle;
   FILEFINDBUF3  FindBuffer;
   FILESTATUS    FileInfoBuf;
   ULONG         FindCount;
   APIRET        rc;           /* Return code */
   HFILE         EPM_handle, our_handle;

   strcpy(edit_cmd, "save");
   if (strlen(filename)) {
      fn_ptr = filename;
   } else {
      PLSTRING lstr_ptr;
      EtkQueryFileField( hwndClient, FILENAME_FIELD, 0 /*Fileid*/, (PLONG)&lstr_ptr);
      memcpy(zFileName, lstr_ptr->Data, lstr_ptr->lsLength);
      zFileName[lstr_ptr->lsLength] = '\0';
      fn_ptr = zFileName;
      strcat(edit_cmd, " ");
      strcat(edit_cmd, fn_ptr);
   } /* endif */

   FindHandle = 0x0001;
   FindCount = 1;

   rc = DosFindFirst(fn_ptr,                /* File pattern */
                     &FindHandle,           /* Directory search handle */
                     0,                     /* Search attribute */
                     (PVOID) &FindBuffer,   /* Result buffer */
                     sizeof(FindBuffer),    /* Result buffer length */
                     &FindCount,            /* # of entries to find */
                     FIL_STANDARD);         /* Return level 1 file info */

   if (rc != 0) {
      sprintf(edit_cmd, "sayerror DosFindFirst error: return code = %ld", rc);
      EtkExecuteCommand( hwndClient, edit_cmd);
      return rc;
   } /* endif */
   rc = EtkExecuteCommand( hwndClient, edit_cmd);
   if (rc != 0) {  /* Error saving; don't continue. */
      return rc;
   } /* endif */
   EtkQueryFileField( hwndClient, LOCKHANDLE_FIELD, 0 /*Fileid*/, (PLONG)&EPM_handle);
   if (!EPM_handle) {  /* If file not locked (Edit /K), we need to open it. */
      ULONG   Action;
      rc = DosOpen(fn_ptr,                  /* File path name */
                   &our_handle,             /* File handle */
                   &Action,                 /* Action taken */
                   0,                       /* File primary allocation */
                   0,                       /* File attribute */
                   OPEN_ACTION_FAIL_IF_NEW |
                   OPEN_ACTION_OPEN_IF_EXISTS,  /* Open function type */
                   OPEN_SHARE_DENYREADWRITE |
                   OPEN_ACCESS_READWRITE,       /* Open mode of the file */
                   0);                      /* No extended attributes */
      if (rc != 0) {
          sprintf(edit_cmd, "sayerror DosOpen error: return code = %ld", rc);
          EtkExecuteCommand( hwndClient, edit_cmd);
          return rc;
      } /* endif */
   } else {
      our_handle = EPM_handle;
   } /* endif */

   rc = DosQueryFileInfo(our_handle, 1, &FileInfoBuf, sizeof(FileInfoBuf) );
                         /* Obtain a copy of the Level 1 */
                         /*   file information           */

   if (rc != 0) {
       sprintf(edit_cmd, "sayerror DosQueryFileInfo error: return code = %ld", rc);
       EtkExecuteCommand( hwndClient, edit_cmd);
      if (!EPM_handle) DosClose(our_handle);
       return rc;
   } /* endif */

   FileInfoBuf.fdateCreation.year =        FindBuffer.fdateCreation.year;
   FileInfoBuf.fdateCreation.month =       FindBuffer.fdateCreation.month;
   FileInfoBuf.fdateCreation.day =         FindBuffer.fdateCreation.day;
   FileInfoBuf.ftimeCreation.hours =       FindBuffer.ftimeCreation.hours;
   FileInfoBuf.ftimeCreation.minutes =     FindBuffer.ftimeCreation.minutes;
   FileInfoBuf.ftimeCreation.twosecs =     FindBuffer.ftimeCreation.twosecs;
   FileInfoBuf.fdateLastAccess.year =      FindBuffer.fdateLastAccess.year;
   FileInfoBuf.fdateLastAccess.month =     FindBuffer.fdateLastAccess.month;
   FileInfoBuf.fdateLastAccess.day =       FindBuffer.fdateLastAccess.day;
   FileInfoBuf.ftimeLastAccess.hours =     FindBuffer.ftimeLastAccess.hours;
   FileInfoBuf.ftimeLastAccess.minutes =   FindBuffer.ftimeLastAccess.minutes;
   FileInfoBuf.ftimeLastAccess.twosecs =   FindBuffer.ftimeLastAccess.twosecs;
   FileInfoBuf.fdateLastWrite.year =       FindBuffer.fdateLastWrite.year;
   FileInfoBuf.fdateLastWrite.month =      FindBuffer.fdateLastWrite.month;
   FileInfoBuf.fdateLastWrite.day =        FindBuffer.fdateLastWrite.day;
   FileInfoBuf.ftimeLastWrite.hours =      FindBuffer.ftimeLastWrite.hours;
   FileInfoBuf.ftimeLastWrite.minutes =    FindBuffer.ftimeLastWrite.minutes;
   FileInfoBuf.ftimeLastWrite.twosecs =    FindBuffer.ftimeLastWrite.twosecs;

   rc = DosSetFileInfo(our_handle, 1, &FileInfoBuf, sizeof(FileInfoBuf) );
   if (rc != 0) {
       sprintf(edit_cmd, "sayerror DosSetFileInfo error: return code = %ld", rc);
       EtkExecuteCommand( hwndClient, edit_cmd);
   } /* endif */

   if (!EPM_handle)           /* If we opened the file ourself, */
      DosClose(our_handle);   /* we must close it. */

   return rc;
}

/* Simple demo using EtkAccessLowLevelData, which returns a pointer    */
/* into the line array at the first line for the specified file.       */
ULONG EXPENTRY test( HWND hwndClient, PSZ string) {
   ULONG this_fid, last_line;
   UCHAR text[200];
   PLSTRING * SubLineArray;
   PLSTRING firstline;
   PLSTRING lastline;
   EtkQueryFileID( hwndClient, &this_fid);
      /* N.B.:  EtkAccessLowLevelData doesn't accept 0 to mean "current file". */
   EtkAccessLowLevelData( hwndClient, this_fid, (PPVOID)&SubLineArray);
   firstline = SubLineArray[0];
   EtkQueryFileField( hwndClient, LAST_FIELD, 0 /*Fileid*/, (PLONG)&last_line);
   lastline = SubLineArray[last_line-1];
   sprintf(text, "  Line 1 = SubLineArray[0] =\n%.*s\n\n  Line %ld = SubLineArray[%ld] =\n%.*s",
           min(firstline->lsLength, 40), firstline->Data,
           last_line,
           last_line - 1,
           min(lastline->lsLength, 40), lastline->Data);
   WinMessageBox(HWND_DESKTOP,
                 hwndClient,
                 text,
                 "EPMCSAMP - test",
                 0L,
                 MB_OK + MB_MOVEABLE);
}

/* Find the next duplicated line.                                      */
LONG EXPENTRY FindDup( HWND hwndClient, PSZ string) {
   ULONG this_fid, this_line, last_line;
   PLSTRING * SubLineArray;
   EtkQueryFileID( hwndClient, &this_fid);
      /* N.B.:  EtkAccessLowLevelData doesn't accept 0 to mean "current file". */
   EtkAccessLowLevelData( hwndClient, this_fid, (PPVOID)&SubLineArray);
   EtkQueryFileField( hwndClient, LINE_FIELD, 0 /*Fileid*/, (PLONG)&this_line);
   EtkQueryFileField( hwndClient, LAST_FIELD, 0 /*Fileid*/, (PLONG)&last_line);
   /* Convert from 1-based to 0-based */
   if (this_line)  this_line--;  /* Except if on line 0, don't decrement. */
   last_line--;
   while (this_line<last_line) {
      if (lstreq_test(SubLineArray[this_line], SubLineArray[this_line+1])) {
         EtkSetFileField( hwndClient, LINE_FIELD, 0 /*Fileid*/, (PVOID)(this_line+1));
         return 0;
      } /* endif */
      this_line++;
   } /* endwhile */
   EtkExecuteCommand( hwndClient, "sayerror Duplicate line not found.");
   return -273;  /* Return "String not found" rc. */
}

/* Highlight all instances of a given string.                           */
ULONG EXPENTRY Highlight( HWND hwndClient, PSZ search_str) {
   SHORT rc;
   UCHAR locate_cmd[1600];
   UCHAR attrib_cmd[100];
   int   num_hits=0;
   ULONG col, len, linenum, attrib_level, modify_count;
   SAVE_POS savepos;

   if (!strlen(search_str)) {  /* Nothing to find? */
      EtkExecuteCommand( hwndClient, "sayerror Missing argument.");
      return (ULONG) -268;  /* "Procedure needs more arguments" */
   } /* endif */

   strcpy(locate_cmd, "xcom l ");
   strcat(locate_cmd, search_str);

   save_pos( hwndClient, &savepos );
   EtkQueryFileField( hwndClient, MODIFY_FIELD, 0 /*Fileid*/, (PLONG)&modify_count);

   while (1) {
      if (rc=EtkExecuteCommand( hwndClient, locate_cmd)) {
         if (num_hits) {
            restore_pos( hwndClient, &savepos );
            EtkQueryFileField( hwndClient, ATTRIBUTE_SUPPORT_LEVEL, 0 /*Fileid*/, (PLONG)&attrib_level);
            if (!(attrib_level&EPM_DM_COLOR))  /* Make sure that color support is on */
               EtkSetFileField( hwndClient, ATTRIBUTE_SUPPORT_LEVEL, 0, (PVOID)(attrib_level|EPM_DM_COLOR));
            EtkSetFileField( hwndClient, MODIFY_FIELD, 0 /*Fileid*/, (PVOID)modify_count);
            sprintf(locate_cmd, "sayerror %ld hit%s.", num_hits, num_hits==1 ? "" : "s");
            EtkExecuteCommand( hwndClient, locate_cmd);
            return num_hits;
         } else {
            return rc;  /* Error; probably -273 == "String not found" */
         } /* endif */
      } /* endif */

      num_hits++;
      EtkGetPMInfo(hwndClient, SEARCHPOS, &col);
      EtkGetPMInfo(hwndClient, LSLENGTH, &len);
      EtkQueryFileField( hwndClient, LINE_FIELD, 0 /*Fileid*/, (PLONG)&linenum);

      sprintf(attrib_cmd, "Insert_attr_val_Pair %ld %ld %ld %ld %ld %ld",
              (ULONG)COLORCLASS,                      /* Class          */
              DEFAULT_HIGHLIGHT_COLOR,                /* Attrib. value  */
              linenum,                                /* First line     */
              linenum,                                /* Last line      */
              col,                                    /* First column   */
              col+len-1);                             /* Last column    */
      EtkExecuteCommand( hwndClient, attrib_cmd);

   /* EtkProcessEditKey( hwndClient, ETK_REPEAT_FIND, 1);     */
   /* 1.  We can't get the RC from a repeat_find op, and      */
   /* 2.  If [E]GREP is used, we might get a hit on the same  */
   /*     string multiple times.  Do another Locate, instead. */
      EtkSetFileField( hwndClient, COL_FIELD, 0, (PVOID)(col+len));  /* Move col. past "hit" */
   } /* endwhile */
}

/* Remove highlighting added by Highlight command.            */
ULONG EXPENTRY UnHighlight( HWND hwndClient, PSZ search_str) {
   ULONG this_fid, last_line, modify_count, rc;
   ULONG AttrCount;
   PLSTRING sp;
   ATTRIBRECTYPE * FirstAttribute;
   ATTRIBRECTYPE * AfterLastAttribute;
   CHAR pBuffer[2000];
   PATTRSTRING pAttrString = (PATTRSTRING)pBuffer ;
   ATTRIBVALUETYPE color;

 #ifdef DEBUG
   UCHAR text[100];
 #endif

   color = (ATTRIBVALUETYPE)atoi(search_str);
   if (!color) {
      color = DEFAULT_HIGHLIGHT_COLOR;
   } /* endif */
   EtkQueryFileField( hwndClient, MODIFY_FIELD, 0 /*Fileid*/, (PLONG)&modify_count);
/* EtkQueryFileID( hwndClient, &this_fid); */
   EtkQueryFileField( hwndClient, LAST_FIELD, 0 /*Fileid*/, (PLONG)&last_line);

   do {
      rc = EtkQueryText(hwndClient,
                        0,  /* fileid */
                        last_line,
                        &sp,
                        &FirstAttribute,
                        &AfterLastAttribute) ;
      if (rc != 0 || !sp) return rc ;
      AttrCount = (AfterLastAttribute - FirstAttribute) ; // Typed pointers, no need to divide by sizeof()
   #ifdef DEBUG
      sprintf(text, "line %ld has %ld attrib(s)\n\nFirst = 0x%lx\nLast = 0x%lx\n%.*s", last_line, AttrCount, FirstAttribute, AfterLastAttribute, sp->lsLength, sp->Data);
      if (MBID_CANCEL ==
          WinMessageBox(HWND_DESKTOP,
                        hwndClient,
                        text,
                        "EPMCSAMP - UnHighlight",
                        0L,
                        MB_OK + MB_MOVEABLE))
         break;
   #endif
      if (AttrCount) {
         ULONG i, j ;
         ULONG Found = 0;
         ATTRIBRECTYPE * pattrArray = NULL ;
         TERMTYPE terminator;

         /* I'm guessing that it will be cheaper to first iterate and maybe */
         /* have to iterate again than to just do it once.  This is because */
         /* I'm guessing that most lines either won't have any attributes,  */
         /* in which case this is trivial, or will have only attributes     */
         /* added by the Highlight command, in which case we'll be removing */
         /* them all, so there's no need to allocate and free a buffer into */
         /* which we copy the ones we want to keep.                   LAM   */
         for (i=0; i<AttrCount; i++) {
            if (FirstAttribute[(USHORT)i].Class == COLORCLASS &&
                FirstAttribute[(USHORT)i].Value == color)
               Found++;
         }
         if (!Found)               /* If no matching attributes,   */
            continue;              /* nothing to do for this line. */

         EtkQueryLineTerminator(hwndClient,
                                0 /* fileid */,
                                last_line,
                                &terminator);

         if (Found==AttrCount) {  /* Delete them all. */
            pAttrString->Attrs  = (ATTRIBRECTYPE*)&pBuffer[0] ;
            pAttrString->ALAttr = (ATTRIBRECTYPE*)&pBuffer[0] ;
         } else {  /* Have to keep some and delete some... */
            pattrArray = (ATTRIBRECTYPE*)malloc((int)(Found)*sizeof(ATTRIBRECTYPE)) ;
            for (i=j=0; i<AttrCount; i++) {
               if (FirstAttribute[(USHORT)i].Class != COLORCLASS ||
                   FirstAttribute[(USHORT)i].Value != color) {
                  memcpy(&pattrArray[(USHORT)j], &FirstAttribute[(USHORT)i], sizeof(ATTRIBRECTYPE)) ;
                  j++ ;
               } /* endif */
            }
            pAttrString->Attrs  = &pattrArray[0] ;
            pAttrString->ALAttr = &pattrArray[j] ;
         } /* endif */

         //----------------------------------------------------------------------
         // Update the current line
         //----------------------------------------------------------------------
         pAttrString->SelfPtr= (CHAR*)&pAttrString->SelfPtr ;
         pAttrString->TextLen= sp->lsLength;
         memcpy(&pAttrString->Text, sp->Data, sp->lsLength);
         EtkReplaceText(hwndClient, 0 /*fid*/, last_line, &pAttrString, terminator);
         if (pattrArray)
            free(pattrArray);
      } /* endif */
   } while ( --last_line ); /* enddo */
   EtkSetFileField( hwndClient, MODIFY_FIELD, 0 /*Fileid*/, (PVOID)modify_count);
}

/*-------------------------- Utility functions --------------------------*/

VOID save_pos( HWND hwndClient, PSAVE_POS save_pos ) {
   EtkQueryFileField( hwndClient, LINE_FIELD,    0 /*Fileid*/, (PLONG)&save_pos->line);
   EtkQueryFileField( hwndClient, COL_FIELD,     0 /*Fileid*/, (PLONG)&save_pos->col );
   EtkQueryFileField( hwndClient, CURSORX_FIELD, 0 /*Fileid*/, (PLONG)&save_pos->cx  );
   EtkQueryFileField( hwndClient, CURSORY_FIELD, 0 /*Fileid*/, (PLONG)&save_pos->cy  );
   #ifdef DEBUG
    {
      unsigned char Text[256];
      PLSTRING pFilename;
      sprintf(Text, "Saved column = %ld\nSaved CX = %ld", save_pos->col, save_pos->cx );
      WinMessageBox(HWND_DESKTOP,
                    hwndClient,
                    Text,
                    "EPMCSAMP - save_pos",
                    0L,
                    MB_OK + MB_MOVEABLE);
    }
   #endif
}

VOID restore_pos( HWND hwndClient, PSAVE_POS save_pos ) {
   ULONG  temp;
   EtkSetFileField( hwndClient, CURSORY_FIELD, 0 /*Fileid*/,  (PVOID)save_pos->cy  );
   EtkQueryFileField( hwndClient, LAST_FIELD, 0 /*Fileid*/, (PLONG)&temp);
   #ifdef DEBUG
    {
      unsigned char Text[256];
      sprintf(Text, ".last = %ld\nSaved line = %ld\min = %ld", temp, save_pos->line, min(temp, save_pos->line));
      WinMessageBox(HWND_DESKTOP,
                    hwndClient,
                    Text,
                    "EPMCSAMP - restore_pos",
                    0L,
                    MB_OK + MB_MOVEABLE);
    }
   #endif
   temp = min(temp, save_pos->line);
   EtkSetFileField( hwndClient, LINE_FIELD,    0 /*Fileid*/,  (PVOID)temp);
   temp = MAXCOL;                        /* Set left edge of window */
   EtkSetFileField( hwndClient, COL_FIELD,     0 /*Fileid*/,  (PVOID)temp );
   temp = save_pos->col - save_pos->cx + 1;
   #ifdef DEBUG
    {
      unsigned char Text[256];
      sprintf(Text, "Saved column = %ld\nSaved CX = %ld\nCol - cx + 1 = %ld", save_pos->col, save_pos->cx, temp);
      WinMessageBox(HWND_DESKTOP,
                    hwndClient,
                    Text,
                    "EPMCSAMP - restore_pos",
                    0L,
                    MB_OK + MB_MOVEABLE);
    }
   #endif
   EtkSetFileField( hwndClient, COL_FIELD,     0 /*Fileid*/,  (PVOID)temp );
                                         /* Set column */
   EtkSetFileField( hwndClient, COL_FIELD,     0 /*Fileid*/,  (PVOID)save_pos->col );
}


#define Length lsLength
BOOL is_substr(PSZ word, PLSTRING line)
{
   register SHORT  wlen;
   register UCHAR  ch;
   register PUCHAR ptr;
   register PUCHAR end;

   if ( (line->Length) && (wlen=strlen(word)) && wlen <= line->Length) {
      ptr = line->Data - 1;
      end = line->Data + line->Length - wlen;
      ch = word[0];
      while (ptr=(PUCHAR)memchr(ptr+1, ch, end-ptr)) {  // Not end-ptr+1, since ptr is 1 below start pos
         if (!memcmp(ptr, word, wlen)) {
            return TRUE;
         } /* endif */
      } /* endwhile */
   } /* endif */
   return FALSE;
}

VOID lstrupcase( PLSTRING s )
{ LSTRINGLENTYPE i;
  for (i=s->Length; i; ){
     i--;
     s->Data[i]=toupper(s->Data[i]);
  }
}

BOOL _Optlink lstreq_test( const PLSTRING s1, const PLSTRING s2)
{
// The following checks both the length and the data, in one shot:
   return(memcmp(s1,s2,s1->Length + LSTRINGLENTYPELEN)==0);
}

ULONG _Optlink EtkInsertAttribute (HWND hwnd, ULONG ulFileId,
                                   ULONG Line, SHORT Col,
                                   ULONG ulAttributeClass, ULONG ulValue, BOOL bPush) {
//------------------------------------------------------------------------------
// This function inserts an attribute at the given position
//------------------------------------------------------------------------------
   ULONG rc ;
   ATTRIBRECTYPE * pattrFirst = NULL ;
   ATTRIBRECTYPE * pattrLast  = NULL ;
   ATTRIBRECTYPE * pattrArray = NULL ;
   ATTRIBRECTYPE * pattrNew   = NULL ;
   PLSTRING pLine = (PLSTRING)0;
   TERMTYPE terminator;
   ULONG ulAttrCount ;
   USHORT usLineLen ;

   rc = EtkQueryText(hwnd,
                     ulFileId,
                     Line,
                     &pLine,
                     &pattrFirst,
                     &pattrLast) ;
   if (rc != 0 || !pLine) return rc ;

   rc = EtkQueryLineTerminator(hwnd,
                               ulFileId,
                               Line,
                               &terminator);

   usLineLen = min(pLine->lsLength, MAXCOL); // Pb in EtkReplaceText when line is too big !!

   //----------------------------------------------------------------------
   // Allocate a new attribute structure containing all the current attributes
   // and a an empty slot for the new attribute
   //----------------------------------------------------------------------
   {
      // We want to know how many attributes we have before we add a new one
      // Note: In a given line, all attributes are stored in pointers so that
      // the following thing makes sense.
      ulAttrCount = (pattrLast - pattrFirst) ; // Typed pointers, no need to divide by sizeof()
      // Allocate enough memory for the existing attributes + one
      pattrArray = (ATTRIBRECTYPE*)malloc((int)(ulAttrCount+1)*sizeof(ATTRIBRECTYPE)) ;
      // Copy the original attributes
      memcpy(pattrArray, pattrFirst, (int)ulAttrCount*sizeof(ATTRIBRECTYPE)) ;
      // Initialize the 'new' attribute
      pattrNew = &pattrArray[ulAttrCount] ;
      // Check that there is no attribute whose column is greater than
      // usCol. If there is one, we must shift a few things
      {
         ULONG i ;
         BOOL bFound = FALSE ;
         for (i=0; i<ulAttrCount; i++) {
            if (pattrArray[(USHORT)i].Col > Col) { bFound = TRUE ; break ;} /* endif */
         }
         if (bFound) {
            // Shift to the right
            memmove(&pattrArray[(USHORT)i+1], &pattrArray[(USHORT)i], (ulAttrCount-i)*sizeof(ATTRIBRECTYPE)) ;
            pattrNew = &pattrArray[(USHORT)i] ;
         } /* endif */
      }
      // Zero the new attr
      memset(pattrNew, 0, sizeof(ATTRIBRECTYPE)) ;
   }
   //----------------------------------------------------------------------
   // Add a new attribute at the end
   //----------------------------------------------------------------------
   {
      pattrNew->Col = Col ;

      pattrNew->Class  = (ATTRIBCLASSTYPE)ulAttributeClass ;
      pattrNew->Value  = ulValue ;
      pattrNew->IsPush = (UCHAR)(bPush) ;
   }

   //----------------------------------------------------------------------
   // Update the current line
   //----------------------------------------------------------------------
   {
      CHAR pBuffer[2000];
      PATTRSTRING pAttrString = (PATTRSTRING)pBuffer ;

      pAttrString->Attrs  = &pattrArray[0] ;
      pAttrString->ALAttr = &pattrArray[ulAttrCount+1] ;
      pAttrString->SelfPtr= (CHAR*)&pAttrString->SelfPtr ;
      pAttrString->TextLen= usLineLen ;
      memcpy(&pAttrString->Text, pLine->Data, usLineLen);
      EtkReplaceText(hwnd, ulFileId, Line, &pAttrString, terminator);
   }
   // Cleanup:
   //----------
   free(pattrArray) ;
   return 0 ;
}
